# Version: 0.19
# pylint: skip-file

"""The Versioneer - like a rocketeer, but for versions."""

import configparser
import errno
import json
import os
import re
import subprocess
import sys


class VersioneerConfig:
    """Container for Versioneer configuration parameters."""


def get_root():
    """Get the project root directory.

    We require that all commands are run from the project root, i.e. the
    directory that contains setup.py, setup.cfg, and versioneer.py .
    """
    root = os.path.realpath(os.path.abspath(os.getcwd()))
    setup_py = os.path.join(root, "setup.py")
    versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        # allow 'python path/to/setup.py COMMAND'
        root = os.path.dirname(os.path.realpath(os.path.abspath(sys.argv[0])))
        setup_py = os.path.join(root, "setup.py")
        versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        err = (
            "Versioneer was unable to run the project root directory. "
            "Versioneer requires setup.py to be executed from "
            "its immediate directory (like 'python setup.py COMMAND'), "
            "or in a way that lets it use sys.argv[0] to find the root "
            "(like 'python path/to/setup.py COMMAND')."
        )
        raise VersioneerBadRootError(err)
    try:
        # Certain runtime workflows (setup.py install/develop in a setuptools
        # tree) execute all dependencies in a single python process, so
        # "versioneer" may be imported multiple times, and python's shared
        # module-import table will cache the first one. So we can't use
        # os.path.dirname(__file__), as that will find whichever
        # versioneer.py was first imported, even in later projects.
        me = os.path.realpath(os.path.abspath(__file__))
        me_dir = os.path.normcase(os.path.splitext(me)[0])
        vsr_dir = os.path.normcase(os.path.splitext(versioneer_py)[0])
        if me_dir != vsr_dir:
            print(
                "Warning: build in %s is using versioneer.py from %s"
                % (os.path.dirname(me), versioneer_py)
            )
    except NameError:
        pass
    return root


def get_config_from_root(root):
    """Read the project setup.cfg file to determine Versioneer config."""
    # This might raise EnvironmentError (if setup.cfg is missing), or
    # configparser.NoSectionError (if it lacks a [versioneer] section), or
    # configparser.NoOptionError (if it lacks "VCS="). See the docstring at
    # the top of versioneer.py for instructions on writing your setup.cfg .
    setup_cfg = os.path.join(root, "setup.cfg")
    parser = configparser.ConfigParser()
    with open(setup_cfg, "r") as f:
        parser.read_file(f)
    VCS = parser.get("versioneer", "VCS")  # mandatory

    def get(parser, name):
        if parser.has_option("versioneer", name):
            return parser.get("versioneer", name)
        return None

    cfg = VersioneerConfig()
    cfg.VCS = VCS
    cfg.style = get(parser, "style") or ""
    cfg.versionfile_source = get(parser, "versionfile_source")
    cfg.versionfile_build = get(parser, "versionfile_build")
    cfg.tag_prefix = get(parser, "tag_prefix")
    if cfg.tag_prefix in ("''", '""'):
        cfg.tag_prefix = ""
    cfg.parentdir_prefix = get(parser, "parentdir_prefix")
    cfg.verbose = get(parser, "verbose")
    return cfg


class NotThisMethod(Exception):
    """Exception raised if a method is not valid for the current scenario."""


# these dictionaries contain VCS-specific tools
LONG_VERSION_PY = {}
HANDLERS = {}


def register_vcs_handler(vcs, method):  # decorator
    """Create decorator to mark a method as the handler of a VCS."""

    def decorate(f):
        """Store f in HANDLERS[vcs][method]."""
        if vcs not in HANDLERS:
            HANDLERS[vcs] = {}
        HANDLERS[vcs][method] = f
        return f

    return decorate


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False, env=None):
    """Call the given command(s)."""
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            dispcmd = str([c] + args)
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen(
                [c] + args,
                cwd=cwd,
                env=env,
                stdout=subprocess.PIPE,
                stderr=(subprocess.PIPE if hide_stderr else None),
            )
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %s" % dispcmd)
                print(e)
            return None, None
    else:
        if verbose:
            print("unable to find command, tried %s" % (commands,))
        return None, None
    stdout = p.communicate()[0].strip().decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %s (error)" % dispcmd)
            print("stdout was %s" % stdout)
        return None, p.returncode
    return stdout, p.returncode


LONG_VERSION_PY[
    "git"
] = r'''
# This file helps to compute a version number in source trees obtained from
# git-archive tarball (such as those provided by githubs download-from-tag
# feature). Distribution tarballs (built by setup.py sdist) and build
# directories (produced by setup.py build) will contain a much shorter file
# that just contains the computed version number.

# This file is released into the public domain. Generated by
# versioneer-0.19 (https://github.com/python-versioneer/python-versioneer)

"""Git implementation of _version.py."""

import errno
import os
import re
import subprocess
import sys


def get_keywords():
    """Get the keywords needed to look up the version information."""
    # these strings will be replaced by git during git-archive.
    # setup.py/versioneer.py will grep for the variable names, so they must
    # each be defined on a line of their own. _version.py will just call
    # get_keywords().
    git_refnames = "%(DOLLAR)sFormat:%%d%(DOLLAR)s"
    git_full = "%(DOLLAR)sFormat:%%H%(DOLLAR)s"
    git_date = "%(DOLLAR)sFormat:%%ci%(DOLLAR)s"
    keywords = {"refnames": git_refnames, "full": git_full, "date": git_date}
    return keywords


class VersioneerConfig:
    """Container for Versioneer configuration parameters."""


def get_config():
    """Create, populate and return the VersioneerConfig() object."""
    # these strings are filled in when 'setup.py versioneer' creates
    # _version.py
    cfg = VersioneerConfig()
    cfg.VCS = "git"
    cfg.style = "%(STYLE)s"
    cfg.tag_prefix = "%(TAG_PREFIX)s"
    cfg.parentdir_prefix = "%(PARENTDIR_PREFIX)s"
    cfg.versionfile_source = "%(VERSIONFILE_SOURCE)s"
    cfg.verbose = False
    return cfg


class NotThisMethod(Exception):
    """Exception raised if a method is not valid for the current scenario."""


LONG_VERSION_PY = {}
HANDLERS = {}


def register_vcs_handler(vcs, method):  # decorator
    """Create decorator to mark a method as the handler of a VCS."""
    def decorate(f):
        """Store f in HANDLERS[vcs][method]."""
        if vcs not in HANDLERS:
            HANDLERS[vcs] = {}
        HANDLERS[vcs][method] = f
        return f
    return decorate


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False,
                env=None):
    """Call the given command(s)."""
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            dispcmd = str([c] + args)
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen([c] + args, cwd=cwd, env=env,
                                 stdout=subprocess.PIPE,
                                 stderr=(subprocess.PIPE if hide_stderr
                                         else None))
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %%s" %% dispcmd)
                print(e)
            return None, None
    else:
        if verbose:
            print("unable to find command, tried %%s" %% (commands,))
        return None, None
    stdout = p.communicate()[0].strip().decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %%s (error)" %% dispcmd)
            print("stdout was %%s" %% stdout)
        return None, p.returncode
    return stdout, p.returncode


def versions_from_parentdir(parentdir_prefix, root, verbose):
    """Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes both
    the project name and a version string. We will also support searching up
    two directory levels for an appropriately named parent directory
    """
    rootdirs = []

    for i in range(3):
        dirname = os.path.basename(root)
        if dirname.startswith(parentdir_prefix):
            return {"version": dirname[len(parentdir_prefix):],
                    "full-revisionid": None,
                    "dirty": False, "error": None, "date": None}
        else:
            rootdirs.append(root)
            root = os.path.dirname(root)  # up a level

    if verbose:
        print("Tried directories %%s but none started with prefix %%s" %%
              (str(rootdirs), parentdir_prefix))
    raise NotThisMethod("rootdir doesn't start with parentdir_prefix")


@register_vcs_handler("git", "get_keywords")
def git_get_keywords(versionfile_abs):
    """Extract version information from the given file."""
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = {}
    try:
        f = open(versionfile_abs, "r")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["full"] = mo.group(1)
            if line.strip().startswith("git_date ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["date"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return keywords


@register_vcs_handler("git", "keywords")
def git_versions_from_keywords(keywords, tag_prefix, verbose):
    """Get version information from git keywords."""
    if not keywords:
        raise NotThisMethod("no keywords at all, weird")
    date = keywords.get("date")
    if date is not None:
        # Use only the last line.  Previous lines may contain GPG signature
        # information.
        date = date.splitlines()[-1]

        # git-2.2.0 added "%%cI", which expands to an ISO-8601 -compliant
        # datestamp. However we prefer "%%ci" (which expands to an "ISO-8601
        # -like" string, which we must then edit to make compliant), because
        # it's been around since git-1.5.3, and it's too difficult to
        # discover which version we're using, or to work around using an
        # older one.
        date = date.strip().replace(" ", "T", 1).replace(" ", "", 1)
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        raise NotThisMethod("unexpanded keywords, not a git-archive tarball")
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %%d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r'\d', r)])
        if verbose:
            print("discarding '%%s', no digits" %% ",".join(refs - tags))
    if verbose:
        print("likely tags: %%s" %% ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking %%s" %% r)
            return {"version": r,
                    "full-revisionid": keywords["full"].strip(),
                    "dirty": False, "error": None,
                    "date": date}
    # no suitable tags, so version is "0+unknown", but full hex is still there
    if verbose:
        print("no suitable tags, using unknown + full revision id")
    return {"version": "0+unknown",
            "full-revisionid": keywords["full"].strip(),
            "dirty": False, "error": "no suitable tags", "date": None}


@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, run_command=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]

    out, rc = run_command(GITS, ["rev-parse", "--git-dir"], cwd=root,
                          hide_stderr=True)
    if rc != 0:
        if verbose:
            print("Directory %%s not under git control" %% root)
        raise NotThisMethod("'git rev-parse --git-dir' returned error")

    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out, rc = run_command(GITS, ["describe", "--tags", "--dirty",
                                          "--always", "--long",
                                          "--match", "%%s*" %% tag_prefix],
                                   cwd=root)
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out, rc = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[:git_describe.rindex("-dirty")]

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r'^(.+)-(\d+)-g([0-9a-f]+)$', git_describe)
        if not mo:
            # unparseable. Maybe git-describe is misbehaving?
            pieces["error"] = ("unable to parse git-describe output: '%%s'"
                               %% describe_out)
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%%s' doesn't start with prefix '%%s'"
                print(fmt %% (full_tag, tag_prefix))
            pieces["error"] = ("tag '%%s' doesn't start with prefix '%%s'"
                               %% (full_tag, tag_prefix))
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix):]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out, rc = run_command(GITS, ["rev-list", "HEAD", "--count"],
                                    cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    # commit date: see ISO-8601 comment in git_versions_from_keywords()
    date = run_command(GITS, ["show", "-s", "--format=%%ci", "HEAD"],
                       cwd=root)[0].strip()
    # Use only the last line.  Previous lines may contain GPG signature
    # information.
    date = date.splitlines()[-1]
    pieces["date"] = date.strip().replace(" ", "T", 1).replace(" ", "", 1)

    return pieces


def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%%d.g%%s" %% (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%%d.g%%s" %% (pieces["distance"],
                                          pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_pre(pieces):
    """TAG[.post0.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post0.devDISTANCE
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".post0.dev%%d" %% pieces["distance"]
    else:
        # exception #1
        rendered = "0.post0.dev%%d" %% pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%%d" %% pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%%s" %% pieces["short"]
    else:
        # exception #1
        rendered = "0.post%%d" %% pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%%s" %% pieces["short"]
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%%d" %% pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%%d" %% pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%%d-g%%s" %% (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%%d-g%%s" %% (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {"version": "unknown",
                "full-revisionid": pieces.get("long"),
                "dirty": None,
                "error": pieces["error"],
                "date": None}

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%%s'" %% style)

    return {"version": rendered, "full-revisionid": pieces["long"],
            "dirty": pieces["dirty"], "error": None,
            "date": pieces.get("date")}


def get_versions():
    """Get version information or return default if unable to do so."""
    # I am in _version.py, which lives at ROOT/VERSIONFILE_SOURCE. If we have
    # __file__, we can work backwards from there to the root. Some
    # py2exe/bbfreeze/non-CPython implementations don't do __file__, in which
    # case we can only use expanded keywords.

    cfg = get_config()
    verbose = cfg.verbose

    try:
        return git_versions_from_keywords(get_keywords(), cfg.tag_prefix,
                                          verbose)
    except NotThisMethod:
        pass

    try:
        root = os.path.realpath(__file__)
        # versionfile_source is the relative path from the top of the source
        # tree (where the .git directory might live) to this file. Invert
        # this to find the root from __file__.
        for i in cfg.versionfile_source.split('/'):
            root = os.path.dirname(root)
    except NameError:
        return {"version": "0+unknown", "full-revisionid": None,
                "dirty": None,
                "error": "unable to find root of source tree",
                "date": None}

    try:
        pieces = git_pieces_from_vcs(cfg.tag_prefix, root, verbose)
        return render(pieces, cfg.style)
    except NotThisMethod:
        pass

    try:
        if cfg.parentdir_prefix:
            return versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
    except NotThisMethod:
        pass

    return {"version": "0+unknown", "full-revisionid": None,
            "dirty": None,
            "error": "unable to compute version", "date": None}
'''


@register_vcs_handler("git", "get_keywords")
def git_get_keywords(versionfile_abs):
    """Extract version information from the given file."""
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = {}
    try:
        f = open(versionfile_abs, "r")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["full"] = mo.group(1)
            if line.strip().startswith("git_date ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["date"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return keywords


@register_vcs_handler("git", "keywords")
def git_versions_from_keywords(keywords, tag_prefix, verbose):
    """Get version information from git keywords."""
    if not keywords:
        raise NotThisMethod("no keywords at all, weird")
    date = keywords.get("date")
    if date is not None:
        # Use only the last line.  Previous lines may contain GPG signature
        # information.
        date = date.splitlines()[-1]

        # git-2.2.0 added "%cI", which expands to an ISO-8601 -compliant
        # datestamp. However we prefer "%ci" (which expands to an "ISO-8601
        # -like" string, which we must then edit to make compliant), because
        # it's been around since git-1.5.3, and it's too difficult to
        # discover which version we're using, or to work around using an
        # older one.
        date = date.strip().replace(" ", "T", 1).replace(" ", "", 1)
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        raise NotThisMethod("unexpanded keywords, not a git-archive tarball")
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG) :] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r"\d", r)])
        if verbose:
            print("discarding '%s', no digits" % ",".join(refs - tags))
    if verbose:
        print("likely tags: %s" % ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix) :]
            if verbose:
                print("picking %s" % r)
            return {
                "version": r,
                "full-revisionid": keywords["full"].strip(),
                "dirty": False,
                "error": None,
                "date": date,
            }
    # no suitable tags, so version is "0+unknown", but full hex is still there
    if verbose:
        print("no suitable tags, using unknown + full revision id")
    return {
        "version": "0+unknown",
        "full-revisionid": keywords["full"].strip(),
        "dirty": False,
        "error": "no suitable tags",
        "date": None,
    }


@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, run_command=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]

    out, rc = run_command(GITS, ["rev-parse", "--git-dir"], cwd=root, hide_stderr=True)
    if rc != 0:
        if verbose:
            print("Directory %s not under git control" % root)
        raise NotThisMethod("'git rev-parse --git-dir' returned error")

    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out, rc = run_command(
        GITS,
        [
            "describe",
            "--tags",
            "--dirty",
            "--always",
            "--long",
            "--match",
            "%s*" % tag_prefix,
        ],
        cwd=root,
    )
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out, rc = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[: git_describe.rindex("-dirty")]

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r"^(.+)-(\d+)-g([0-9a-f]+)$", git_describe)
        if not mo:
            # unparseable. Maybe git-describe is misbehaving?
            pieces["error"] = "unable to parse git-describe output: '%s'" % describe_out
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%s' doesn't start with prefix '%s'"
                print(fmt % (full_tag, tag_prefix))
            pieces["error"] = "tag '%s' doesn't start with prefix '%s'" % (
                full_tag,
                tag_prefix,
            )
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix) :]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out, rc = run_command(GITS, ["rev-list", "HEAD", "--count"], cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    # commit date: see ISO-8601 comment in git_versions_from_keywords()
    date = run_command(GITS, ["show", "-s", "--format=%ci", "HEAD"], cwd=root)[
        0
    ].strip()
    # Use only the last line.  Previous lines may contain GPG signature
    # information.
    date = date.splitlines()[-1]
    pieces["date"] = date.strip().replace(" ", "T", 1).replace(" ", "", 1)

    return pieces


def do_vcs_install(manifest_in, versionfile_source, ipy):
    """Git-specific installation logic for Versioneer.

    For Git, this means creating/changing .gitattributes to mark _version.py
    for export-subst keyword substitution.
    """
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    files = [manifest_in, versionfile_source]
    if ipy:
        files.append(ipy)
    try:
        me = __file__
        if me.endswith(".pyc") or me.endswith(".pyo"):
            me = os.path.splitext(me)[0] + ".py"
        versioneer_file = os.path.relpath(me)
    except NameError:
        versioneer_file = "versioneer.py"
    files.append(versioneer_file)
    present = False
    try:
        f = open(".gitattributes", "r")
        for line in f.readlines():
            if line.strip().startswith(versionfile_source):
                if "export-subst" in line.strip().split()[1:]:
                    present = True
        f.close()
    except EnvironmentError:
        pass
    if not present:
        f = open(".gitattributes", "a+")
        f.write("%s export-subst\n" % versionfile_source)
        f.close()
        files.append(".gitattributes")
    run_command(GITS, ["add", "--"] + files)


def versions_from_parentdir(parentdir_prefix, root, verbose):
    """Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes both
    the project name and a version string. We will also support searching up
    two directory levels for an appropriately named parent directory
    """
    rootdirs = []

    for i in range(3):
        dirname = os.path.basename(root)
        if dirname.startswith(parentdir_prefix):
            return {
                "version": dirname[len(parentdir_prefix) :],
                "full-revisionid": None,
                "dirty": False,
                "error": None,
                "date": None,
            }
        else:
            rootdirs.append(root)
            root = os.path.dirname(root)  # up a level

    if verbose:
        print(
            "Tried directories %s but none started with prefix %s"
            % (str(rootdirs), parentdir_prefix)
        )
    raise NotThisMethod("rootdir doesn't start with parentdir_prefix")


SHORT_VERSION_PY = """
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
%s
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
"""


def versions_from_file(filename):
    """Try to determine the version from _version.py if present."""
    try:
        with open(filename) as f:
            contents = f.read()
    except EnvironmentError:
        raise NotThisMethod("unable to read _version.py")
    mo = re.search(
        r"version_json = '''\n(.*)'''  # END VERSION_JSON", contents, re.M | re.S
    )
    if not mo:
        mo = re.search(
            r"version_json = '''\r\n(.*)'''  # END VERSION_JSON", contents, re.M | re.S
        )
    if not mo:
        raise NotThisMethod("no version_json in _version.py")
    return json.loads(mo.group(1))


def write_to_version_file(filename, versions):
    """Write the given version number to the given _version.py file."""
    os.unlink(filename)
    contents = json.dumps(versions, sort_keys=True, indent=1, separators=(",", ": "))
    with open(filename, "w") as f:
        f.write(SHORT_VERSION_PY % contents)

    print("set %s to '%s'" % (filename, versions["version"]))


def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%d.g%s" % (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%d.g%s" % (pieces["distance"], pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_pre(pieces):
    """TAG[.post0.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post0.devDISTANCE
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".post0.dev%d" % pieces["distance"]
    else:
        # exception #1
        rendered = "0.post0.dev%d" % pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%s" % pieces["short"]
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {
            "version": "unknown",
            "full-revisionid": pieces.get("long"),
            "dirty": None,
            "error": pieces["error"],
            "date": None,
        }

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%s'" % style)

    return {
        "version": rendered,
        "full-revisionid": pieces["long"],
        "dirty": pieces["dirty"],
        "error": None,
        "date": pieces.get("date"),
    }


class VersioneerBadRootError(Exception):
    """The project root directory is unknown or missing key files."""


def get_versions(verbose=False):
    """Get the project version from whatever source is available.

    Returns dict with two keys: 'version' and 'full'.
    """
    if "versioneer" in sys.modules:
        # see the discussion in cmdclass.py:get_cmdclass()
        del sys.modules["versioneer"]

    root = get_root()
    cfg = get_config_from_root(root)

    assert cfg.VCS is not None, "please set [versioneer]VCS= in setup.cfg"
    handlers = HANDLERS.get(cfg.VCS)
    assert handlers, "unrecognized VCS '%s'" % cfg.VCS
    verbose = verbose or cfg.verbose
    assert (
        cfg.versionfile_source is not None
    ), "please set versioneer.versionfile_source"
    assert cfg.tag_prefix is not None, "please set versioneer.tag_prefix"

    versionfile_abs = os.path.join(root, cfg.versionfile_source)

    # extract version from first of: _version.py, VCS command (e.g. 'git
    # describe'), parentdir. This is meant to work for developers using a
    # source checkout, for users of a tarball created by 'setup.py sdist',
    # and for users of a tarball/zipball created by 'git archive' or github's
    # download-from-tag feature or the equivalent in other VCSes.

    get_keywords_f = handlers.get("get_keywords")
    from_keywords_f = handlers.get("keywords")
    if get_keywords_f and from_keywords_f:
        try:
            keywords = get_keywords_f(versionfile_abs)
            ver = from_keywords_f(keywords, cfg.tag_prefix, verbose)
            if verbose:
                print("got version from expanded keyword %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        ver = versions_from_file(versionfile_abs)
        if verbose:
            print("got version from file %s %s" % (versionfile_abs, ver))
        return ver
    except NotThisMethod:
        pass

    from_vcs_f = handlers.get("pieces_from_vcs")
    if from_vcs_f:
        try:
            pieces = from_vcs_f(cfg.tag_prefix, root, verbose)
            ver = render(pieces, cfg.style)
            if verbose:
                print("got version from VCS %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        if cfg.parentdir_prefix:
            ver = versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
            if verbose:
                print("got version from parentdir %s" % ver)
            return ver
    except NotThisMethod:
        pass

    if verbose:
        print("unable to compute version")

    return {
        "version": "0+unknown",
        "full-revisionid": None,
        "dirty": None,
        "error": "unable to compute version",
        "date": None,
    }


def get_version():
    """Get the short version string for this project."""
    return get_versions()["version"]


def get_cmdclass(cmdclass=None):
    """Get the custom setuptools/distutils subclasses used by Versioneer.

    If the package uses a different cmdclass (e.g. one from numpy), it
    should be provide as an argument.
    """
    if "versioneer" in sys.modules:
        del sys.modules["versioneer"]
        # this fixes the "python setup.py develop" case (also 'install' and
        # 'easy_install .'), in which subdependencies of the main project are
        # built (using setup.py bdist_egg) in the same python process. Assume
        # a main project A and a dependency B, which use different versions
        # of Versioneer. A's setup.py imports A's Versioneer, leaving it in
        # sys.modules by the time B's setup.py is executed, causing B to run
        # with the wrong versioneer. Setuptools wraps the sub-dep builds in a
        # sandbox that restores sys.modules to it's pre-build state, so the
        # parent is protected against the child's "import versioneer". By
        # removing ourselves from sys.modules here, before the child build
        # happens, we protect the child from the parent's versioneer too.
        # Also see https://github.com/python-versioneer/python-versioneer/issues/52

    cmds = {} if cmdclass is None else cmdclass.copy()

    # we add "version" to both distutils and setuptools
    from distutils.core import Command

    class cmd_version(Command):
        description = "report generated version string"
        user_options = []
        boolean_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            vers = get_versions(verbose=True)
            print("Version: %s" % vers["version"])
            print(" full-revisionid: %s" % vers.get("full-revisionid"))
            print(" dirty: %s" % vers.get("dirty"))
            print(" date: %s" % vers.get("date"))
            if vers["error"]:
                print(" error: %s" % vers["error"])

    cmds["version"] = cmd_version

    # we override "build_py" in both distutils and setuptools
    #
    # most invocation pathways end up running build_py:
    #  distutils/build -> build_py
    #  distutils/install -> distutils/build ->..
    #  setuptools/bdist_wheel -> distutils/install ->..
    #  setuptools/bdist_egg -> distutils/install_lib -> build_py
    #  setuptools/install -> bdist_egg ->..
    #  setuptools/develop -> ?
    #  pip install:
    #   copies source tree to a tempdir before running egg_info/etc
    #   if .git isn't copied too, 'git describe' will fail
    #   then does setup.py bdist_wheel, or sometimes setup.py install
    #  setup.py egg_info -> ?

    # we override different "build_py" commands for both environments
    if "build_py" in cmds:
        _build_py = cmds["build_py"]
    elif "setuptools" in sys.modules:
        from setuptools.command.build_py import build_py as _build_py
    else:
        from distutils.command.build_py import build_py as _build_py

    class cmd_build_py(_build_py):
        def run(self):
            root = get_root()
            cfg = get_config_from_root(root)
            versions = get_versions()
            _build_py.run(self)
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            if cfg.versionfile_build:
                target_versionfile = os.path.join(self.build_lib, cfg.versionfile_build)
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

    cmds["build_py"] = cmd_build_py

    if "setuptools" in sys.modules:
        from setuptools.command.build_ext import build_ext as _build_ext
    else:
        from distutils.command.build_ext import build_ext as _build_ext

    class cmd_build_ext(_build_ext):
        def run(self):
            root = get_root()
            cfg = get_config_from_root(root)
            versions = get_versions()
            _build_ext.run(self)
            if self.inplace:
                # build_ext --inplace will only build extensions in
                # build/lib<..> dir with no _version.py to write to.
                # As in place builds will already have a _version.py
                # in the module dir, we do not need to write one.
                return
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            target_versionfile = os.path.join(self.build_lib, cfg.versionfile_source)
            print("UPDATING %s" % target_versionfile)
            write_to_version_file(target_versionfile, versions)

    cmds["build_ext"] = cmd_build_ext

    if "cx_Freeze" in sys.modules:  # cx_freeze enabled?
        from cx_Freeze.dist import build_exe as _build_exe

        # nczeczulin reports that py2exe won't like the pep440-style string
        # as FILEVERSION, but it can be used for PRODUCTVERSION, e.g.
        # setup(console=[{
        #   "version": versioneer.get_version().split("+", 1)[0], # FILEVERSION
        #   "product_version": versioneer.get_version(),
        #   ...

        class cmd_build_exe(_build_exe):
            def run(self):
                root = get_root()
                cfg = get_config_from_root(root)
                versions = get_versions()
                target_versionfile = cfg.versionfile_source
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

                _build_exe.run(self)
                os.unlink(target_versionfile)
                with open(cfg.versionfile_source, "w") as f:
                    LONG = LONG_VERSION_PY[cfg.VCS]
                    f.write(
                        LONG
                        % {
                            "DOLLAR": "$",
                            "STYLE": cfg.style,
                            "TAG_PREFIX": cfg.tag_prefix,
                            "PARENTDIR_PREFIX": cfg.parentdir_prefix,
                            "VERSIONFILE_SOURCE": cfg.versionfile_source,
                        }
                    )

        cmds["build_exe"] = cmd_build_exe
        del cmds["build_py"]

    if "py2exe" in sys.modules:  # py2exe enabled?
        from py2exe.distutils_buildexe import py2exe as _py2exe

        class cmd_py2exe(_py2exe):
            def run(self):
                root = get_root()
                cfg = get_config_from_root(root)
                versions = get_versions()
                target_versionfile = cfg.versionfile_source
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

                _py2exe.run(self)
                os.unlink(target_versionfile)
                with open(cfg.versionfile_source, "w") as f:
                    LONG = LONG_VERSION_PY[cfg.VCS]
                    f.write(
                        LONG
                        % {
                            "DOLLAR": "$",
                            "STYLE": cfg.style,
                            "TAG_PREFIX": cfg.tag_prefix,
                            "PARENTDIR_PREFIX": cfg.parentdir_prefix,
                            "VERSIONFILE_SOURCE": cfg.versionfile_source,
                        }
                    )

        cmds["py2exe"] = cmd_py2exe

    # we override different "sdist" commands for both environments
    if "sdist" in cmds:
        _sdist = cmds["sdist"]
    elif "setuptools" in sys.modules:
        from setuptools.command.sdist import sdist as _sdist
    else:
        from distutils.command.sdist import sdist as _sdist

    class cmd_sdist(_sdist):
        def run(self):
            versions = get_versions()
            self._versioneer_generated_versions = versions
            # unless we update this, the command will keep using the old
            # version
            self.distribution.metadata.version = versions["version"]
            return _sdist.run(self)

        def make_release_tree(self, base_dir, files):
            root = get_root()
            cfg = get_config_from_root(root)
            _sdist.make_release_tree(self, base_dir, files)
            # now locate _version.py in the new base_dir directory
            # (remembering that it may be a hardlink) and replace it with an
            # updated value
            target_versionfile = os.path.join(base_dir, cfg.versionfile_source)
            print("UPDATING %s" % target_versionfile)
            write_to_version_file(
                target_versionfile, self._versioneer_generated_versions
            )

    cmds["sdist"] = cmd_sdist

    return cmds


CONFIG_ERROR = """
setup.cfg is missing the necessary Versioneer configuration. You need
a section like:

 [versioneer]
 VCS = git
 style = pep440
 versionfile_source = src/myproject/_version.py
 versionfile_build = myproject/_version.py
 tag_prefix =
 parentdir_prefix = myproject-

You will also need to edit your setup.py to use the results:

 import versioneer
 setup(version=versioneer.get_version(),
       cmdclass=versioneer.get_cmdclass(), ...)

Please read the docstring in ./versioneer.py for configuration instructions,
edit setup.cfg, and re-run the installer or 'python versioneer.py setup'.
"""

SAMPLE_CONFIG = """
# See the docstring in versioneer.py for instructions. Note that you must
# re-run 'versioneer.py setup' after changing this section, and commit the
# resulting files.

[versioneer]
#VCS = git
#style = pep440
#versionfile_source =
#versionfile_build =
#tag_prefix =
#parentdir_prefix =

"""

INIT_PY_SNIPPET = """
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
"""


def do_setup():
    """Do main VCS-independent setup function for installing Versioneer."""
    root = get_root()
    try:
        cfg = get_config_from_root(root)
    except (
        EnvironmentError,
        configparser.NoSectionError,
        configparser.NoOptionError,
    ) as e:
        if isinstance(e, (EnvironmentError, configparser.NoSectionError)):
            print("Adding sample versioneer config to setup.cfg", file=sys.stderr)
            with open(os.path.join(root, "setup.cfg"), "a") as f:
                f.write(SAMPLE_CONFIG)
        print(CONFIG_ERROR, file=sys.stderr)
        return 1

    print(" creating %s" % cfg.versionfile_source)
    with open(cfg.versionfile_source, "w") as f:
        LONG = LONG_VERSION_PY[cfg.VCS]
        f.write(
            LONG
            % {
                "DOLLAR": "$",
                "STYLE": cfg.style,
                "TAG_PREFIX": cfg.tag_prefix,
                "PARENTDIR_PREFIX": cfg.parentdir_prefix,
                "VERSIONFILE_SOURCE": cfg.versionfile_source,
            }
        )

    ipy = os.path.join(os.path.dirname(cfg.versionfile_source), "__init__.py")
    if os.path.exists(ipy):
        try:
            with open(ipy, "r") as f:
                old = f.read()
        except EnvironmentError:
            old = ""
        if INIT_PY_SNIPPET not in old:
            print(" appending to %s" % ipy)
            with open(ipy, "a") as f:
                f.write(INIT_PY_SNIPPET)
        else:
            print(" %s unmodified" % ipy)
    else:
        print(" %s doesn't exist, ok" % ipy)
        ipy = None

    # Make sure both the top-level "versioneer.py" and versionfile_source
    # (PKG/_version.py, used by runtime code) are in MANIFEST.in, so
    # they'll be copied into source distributions. Pip won't be able to
    # install the package without this.
    manifest_in = os.path.join(root, "MANIFEST.in")
    simple_includes = set()
    try:
        with open(manifest_in, "r") as f:
            for line in f:
                if line.startswith("include "):
                    for include in line.split()[1:]:
                        simple_includes.add(include)
    except EnvironmentError:
        pass
    # That doesn't cover everything MANIFEST.in can do
    # (http://docs.python.org/2/distutils/sourcedist.html#commands), so
    # it might give some false negatives. Appending redundant 'include'
    # lines is safe, though.
    if "versioneer.py" not in simple_includes:
        print(" appending 'versioneer.py' to MANIFEST.in")
        with open(manifest_in, "a") as f:
            f.write("include versioneer.py\n")
    else:
        print(" 'versioneer.py' already in MANIFEST.in")
    if cfg.versionfile_source not in simple_includes:
        print(
            " appending versionfile_source ('%s') to MANIFEST.in"
            % cfg.versionfile_source
        )
        with open(manifest_in, "a") as f:
            f.write("include %s\n" % cfg.versionfile_source)
    else:
        print(" versionfile_source already in MANIFEST.in")

    # Make VCS-specific changes. For git, this means creating/changing
    # .gitattributes to mark _version.py for export-subst keyword
    # substitution.
    do_vcs_install(manifest_in, cfg.versionfile_source, ipy)
    return 0


def scan_setup_py():
    """Validate the contents of setup.py against Versioneer's expectations."""
    found = set()
    setters = False
    errors = 0
    with open("setup.py", "r") as f:
        for line in f.readlines():
            if "import versioneer" in line:
                found.add("import")
            if "versioneer.get_cmdclass()" in line:
                found.add("cmdclass")
            if "versioneer.get_version()" in line:
                found.add("get_version")
            if "versioneer.VCS" in line:
                setters = True
            if "versioneer.versionfile_source" in line:
                setters = True
    if len(found) != 3:
        print("")
        print("Your setup.py appears to be missing some important items")
        print("(but I might be wrong). Please make sure it has something")
        print("roughly like the following:")
        print("")
        print(" import versioneer")
        print(" setup( version=versioneer.get_version(),")
        print("        cmdclass=versioneer.get_cmdclass(),  ...)")
        print("")
        errors += 1
    if setters:
        print("You should remove lines like 'versioneer.VCS = ' and")
        print("'versioneer.versionfile_source = ' . This configuration")
        print("now lives in setup.cfg, and should be removed from setup.py")
        print("")
        errors += 1
    return errors


if __name__ == "__main__":
    cmd = sys.argv[1]
    if cmd == "setup":
        errors = do_setup()
        errors += scan_setup_py()
        if errors:
            sys.exit(1)



"""adds GPC info to consent table

Revision ID: d65bbc647083
Revises: c9ee230fa6da
Create Date: 2023-02-15 13:57:36.159161

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d65bbc647083"
down_revision = "c9ee230fa6da"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("consent", sa.Column("has_gpc_flag", sa.Boolean(), nullable=True))
    op.add_column(
        "consent", sa.Column("conflicts_with_gpc", sa.Boolean(), nullable=True)
    )
    op.execute("UPDATE consent SET has_gpc_flag = false")
    op.alter_column(
        "consent", "has_gpc_flag", server_default="f", default=False, nullable=False
    )
    op.execute("UPDATE consent SET conflicts_with_gpc = false")
    op.alter_column(
        "consent",
        "conflicts_with_gpc",
        server_default="f",
        default=False,
        nullable=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("consent", "conflicts_with_gpc")
    op.drop_column("consent", "has_gpc_flag")
    # ### end Alembic commands ###



"""add saas config to connection config

Revision ID: 1dfc5a2d30e7
Revises: e55a51b354e3
Create Date: 2022-02-09 23:27:24.742938

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "1dfc5a2d30e7"
down_revision = "e55a51b354e3"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "connectionconfig",
        sa.Column(
            "saas_config", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )

    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'saas')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    op.drop_column("connectionconfig", "saas_config")
    op.execute("delete from connectionconfig where connection_type in ('saas')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



"""remove data_qualifier

Revision ID: 548a1ac26771
Revises: 848a8f4125cf
Create Date: 2023-12-08 22:07:38.617038

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "548a1ac26771"
down_revision = "848a8f4125cf"
branch_labels = None
depends_on = None


def upgrade():
    ## Remove data qualifier references
    op.drop_index("ix_ctl_data_qualifiers_fides_key", table_name="ctl_data_qualifiers")
    op.drop_index("ix_ctl_data_qualifiers_id", table_name="ctl_data_qualifiers")
    op.drop_table("ctl_data_qualifiers")
    op.drop_column("ctl_datasets", "data_qualifier")
    op.drop_index("ix_ctl_systems_name", table_name="ctl_systems")
    op.drop_column("privacydeclaration", "data_qualifier")

    ## Remove registry references
    op.drop_index("ix_ctl_registries_fides_key", table_name="ctl_registries")
    op.drop_index("ix_ctl_registries_id", table_name="ctl_registries")
    op.drop_table("ctl_registries")
    op.drop_column("ctl_systems", "registry_id")

    ## Remove data use fields
    op.drop_column("ctl_data_uses", "legal_basis")
    op.drop_column("ctl_data_uses", "special_category")
    op.drop_column("ctl_data_uses", "legitimate_interest")
    op.drop_column("ctl_data_uses", "legitimate_interest_impact_assessment")
    op.drop_column("ctl_data_uses", "recipients")

    ## Remove dataset fields
    op.drop_column("ctl_datasets", "retention")
    op.drop_column("ctl_datasets", "joint_controller")
    op.drop_column("ctl_datasets", "third_country_transfers")

    ## Remove system fields
    op.drop_column("ctl_systems", "joint_controller")
    op.drop_column("ctl_systems", "data_responsibility_title")
    op.drop_column("ctl_systems", "third_country_transfers")
    op.drop_column("ctl_systems", "data_protection_impact_assessment")


def downgrade():
    ## Add back in data qualifier references
    op.add_column(
        "privacydeclaration",
        sa.Column("data_qualifier", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_index("ix_ctl_systems_name", "ctl_systems", ["name"], unique=False)
    op.add_column(
        "ctl_datasets",
        sa.Column("data_qualifier", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_table(
        "ctl_data_qualifiers",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column("fides_key", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "organization_fides_key", sa.TEXT(), autoincrement=False, nullable=True
        ),
        sa.Column("name", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("description", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("parent_key", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "tags", postgresql.ARRAY(sa.VARCHAR()), autoincrement=False, nullable=True
        ),
        sa.Column("is_default", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column(
            "active",
            sa.BOOLEAN(),
            server_default=sa.text("true"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("version_added", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("version_deprecated", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("replaced_by", sa.TEXT(), autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint("id", name="ctl_data_qualifiers_pkey"),
    )
    op.create_index(
        "ix_ctl_data_qualifiers_id", "ctl_data_qualifiers", ["id"], unique=False
    )
    op.create_index(
        "ix_ctl_data_qualifiers_fides_key",
        "ctl_data_qualifiers",
        ["fides_key"],
        unique=False,
    )

    ## Add back in registry references
    op.add_column(
        "ctl_systems",
        sa.Column("registry_id", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_table(
        "ctl_registries",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column("fides_key", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "organization_fides_key", sa.TEXT(), autoincrement=False, nullable=True
        ),
        sa.Column("name", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column("description", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "tags", postgresql.ARRAY(sa.VARCHAR()), autoincrement=False, nullable=True
        ),
        sa.PrimaryKeyConstraint("id", name="ctl_registries_pkey"),
    )
    op.create_index("ix_ctl_registries_id", "ctl_registries", ["id"], unique=False)
    op.create_index(
        "ix_ctl_registries_fides_key", "ctl_registries", ["fides_key"], unique=False
    )

    ## Add back system fields
    op.add_column(
        "ctl_systems",
        sa.Column(
            "data_protection_impact_assessment",
            postgresql.JSON(astext_type=sa.Text()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "third_country_transfers",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "data_responsibility_title",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "joint_controller", postgresql.BYTEA(), autoincrement=False, nullable=True
        ),
    )

    ## Add back dataset fields
    op.add_column(
        "ctl_datasets",
        sa.Column(
            "third_country_transfers",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_datasets",
        sa.Column(
            "joint_controller", postgresql.BYTEA(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "ctl_datasets",
        sa.Column("retention", sa.VARCHAR(), autoincrement=False, nullable=True),
    )

    ## Add back data use fields
    op.add_column(
        "ctl_data_uses",
        sa.Column(
            "recipients",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_data_uses",
        sa.Column(
            "legitimate_interest_impact_assessment",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_data_uses",
        sa.Column(
            "legitimate_interest", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "ctl_data_uses",
        sa.Column("special_category", sa.TEXT(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "ctl_data_uses",
        sa.Column("legal_basis", sa.TEXT(), autoincrement=False, nullable=True),
    )



"""Redo roles scopes mapping

Revision ID: 9f38dad37628
Revises: eb1e6ec39b83
Create Date: 2023-03-03 16:53:03.553992

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import ResultProxy

revision = "9f38dad37628"
down_revision = "eb1e6ec39b83"
branch_labels = None
depends_on = None


def swap_roles(old_role: str, new_role: str):
    bind = op.get_bind()
    matched_rows: ResultProxy = bind.execute(
        text("SELECT * FROM fidesuserpermissions WHERE :old_role = ANY(roles);"),
        {"old_role": old_role},
    )
    for row in matched_rows:
        current_roles = row["roles"]
        current_roles.remove(old_role)
        current_roles.append(new_role)
        bind.execute(
            text("UPDATE fidesuserpermissions SET roles= :roles WHERE id= :id"),
            {"roles": sorted(list(set(current_roles))), "id": row["id"]},
        )

    matched_client_rows: ResultProxy = bind.execute(
        text("SELECT * FROM client WHERE :old_role = ANY(roles);"),
        {"old_role": old_role},
    )
    for row in matched_client_rows:
        client_roles = row["roles"]
        client_roles.remove(old_role)
        client_roles.append(new_role)
        bind.execute(
            text("UPDATE client SET roles= :roles WHERE id= :id"),
            {"roles": sorted(list(set(client_roles))), "id": row["id"]},
        )


def upgrade():
    swap_roles("admin", "owner")
    swap_roles("viewer_and_privacy_request_manager", "viewer_and_approver")
    swap_roles("privacy_request_manager", "approver")


def downgrade():
    swap_roles("owner", "admin")
    swap_roles("viewer_and_approver", "viewer_and_privacy_request_manager")
    swap_roles("approver", "privacy_request_manager")



from typing import List, Optional, Type

from sqlalchemy import and_, func, or_
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.selectable import Select

from fides.api.models.sql_models import FidesBase  # type: ignore[attr-defined]
from fides.api.schemas.filter_params import FilterParams


class MissingTaxonomyField(ValueError):
    pass


# FIXME: this code is basically the same as the one in filter_datamap_query
# in the fidesplus repo, but slightly more generic. Ideally we want to replace that with using this
# so we don't duplicate this logic in two different places
def apply_filters_to_query(
    query: Select,
    filter_params: FilterParams,
    search_model: Type[FidesBase],  # Model to search on
    taxonomy_model: Optional[
        Type[FidesBase]
    ],  # Model that has the taxonomy fields to filter on
) -> Select:
    """
    Function to filter a given query by given filter params.
    The search term is used as a filter on the search_model name and fides_key, as well as its id.
    Taxonomy filters are applied to the taxonomy_model if provided.
    The search_model and taxonomy_model may be the same model, e.g if the lookup is on one table,
    or may be different, e.g if the query is performing a join between two tables.
    Returns the filtered query.
    """

    # Perform a text search on the search_model's name, fides_key and id
    if filter_params.search:
        query = query.where(
            and_(
                or_(
                    func.lower(search_model.name).like(
                        f"%{filter_params.search.lower()}%"
                    ),
                    search_model.fides_key == filter_params.search,
                    search_model.id == filter_params.search,
                )
            )
        )

    if not taxonomy_model:
        return query

    # We match the name of the field in FilterParams to the name of the field in the taxonomy_model,
    # which can be represented by either a single element field or a collection field
    taxonomy_field_information = {
        "data_categories": {
            "single": "data_category",
            "collection": "data_categories",
        },
        "data_subjects": {
            "single": "data_subject",
            "collection": "data_subjects",
        },
        "data_uses": {
            "single": "data_use",
            "collection": "data_uses",
        },
    }

    # Filter the fields so we only use the ones that have been provided in the filter params
    available_fields_info = {
        field: field_info
        for field, field_info in taxonomy_field_information.items()
        if getattr(filter_params, field)
    }

    taxonomy_filter_conditions: List[BooleanClauseList] = []

    for field, field_info in available_fields_info.items():
        single_field_name = field_info["single"]
        collection_field_name = field_info["collection"]

        # If the taxonomy_model doesn't have either a single or collection field matching this field
        # we raise an error since it makes no sense to pass in the field as part of the filter params
        if not hasattr(taxonomy_model, single_field_name) and not hasattr(
            taxonomy_model, collection_field_name
        ):
            raise MissingTaxonomyField(
                f"Model {taxonomy_model.__name__} does not have a {single_field_name} or {collection_field_name} field, but filter_params.{field} is not empty"
            )

        single_field_conditions = []
        collection_field_conditions = []

        # For single fields, we match each element provided in the filter params field
        # against the field in the taxonomy model using like, since model field is a single element
        # e.g a single data category represented as a string
        if hasattr(taxonomy_model, single_field_name):
            single_field_conditions = [
                getattr(taxonomy_model, single_field_name).like(element + "%")
                for element in getattr(filter_params, field)
            ]

        # For collection fields, we match each element provided in the filter params field
        # against the field in the taxonomy model using contains, since model field is
        # a collection of elements, e.g a list of data categories
        if hasattr(taxonomy_model, collection_field_name):
            collection_field_conditions = [
                getattr(taxonomy_model, collection_field_name).contains([element])
                for element in getattr(filter_params, field)
            ]

        # We join all conditions with an OR, so that we retrieve rows that match
        # either in their single or collection fields
        all_field_conditions = or_(
            *single_field_conditions, *collection_field_conditions
        )
        taxonomy_filter_conditions.append(all_field_conditions)

    # Finally, we filter the query for taxonomy_model instances that match all the conditions
    if taxonomy_filter_conditions:
        query = query.where(and_(*taxonomy_filter_conditions))

    return query



from typing import Any, Dict, List

from sqlalchemy.orm import Session

from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.models.policy import Rule
from fides.api.models.privacy_request import ExecutionLog, PrivacyRequest
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    AdvancedSettings,
    BaseEmailSchema,
    IdentityTypes,
)
from fides.api.schemas.messaging.messaging import (
    ErasureRequestBodyParams,
    MessagingActionType,
)
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import ExecutionLogStatus
from fides.api.schemas.redis_cache import Identity
from fides.api.service.connectors.base_email_connector import (
    BaseEmailConnector,
    get_org_name,
)
from fides.api.service.messaging.message_dispatch_service import (
    dispatch_message,
    get_email_messaging_config_service_type,
)
from fides.config import get_config

CONFIG = get_config()

ERASURE_EMAIL_CONNECTOR_TYPES = [
    ConnectionType.generic_erasure_email,
    ConnectionType.attentive_email,
    ConnectionType.dynamic_erasure_email,
]


class BaseErasureEmailConnector(BaseEmailConnector):
    """Generic Email Erasure Connector that can be overridden for specific vendors"""

    def get_config(self, configuration: ConnectionConfig) -> BaseEmailSchema:
        raise NotImplementedError("Get config method must be implemented in a subclass")

    @property
    def identities_for_test_email(self) -> Dict[str, Any]:
        return {"email": "test_email@example.com"}

    @property
    def required_identities(self) -> List[str]:
        return get_identity_types_for_connector(self.config)

    def __init__(self, configuration: ConnectionConfig):
        super().__init__(configuration)
        self.config = self.get_config(configuration)

    def needs_email(
        self, user_identities: Dict[str, Any], privacy_request: PrivacyRequest
    ) -> bool:
        """Schedules an erasure email for erasure privacy requests containing the required identities"""
        erasure_rules: List[Rule] = privacy_request.policy.get_rules_for_action(
            action_type=ActionType.erasure
        )
        return bool(
            erasure_rules
            and filter_user_identities_for_connector(self.config, user_identities)
        )

    def add_skipped_log(self, db: Session, privacy_request: PrivacyRequest) -> None:
        """Add skipped log for the email connector to the privacy request"""
        ExecutionLog.create(
            db=db,
            data={
                "connection_key": self.configuration.key,
                "dataset_name": self.configuration.name_or_key,
                "collection_name": self.configuration.name_or_key,
                "privacy_request_id": privacy_request.id,
                "action_type": ActionType.erasure,
                "status": ExecutionLogStatus.skipped,
                "message": f"Erasure email skipped for '{self.configuration.name_or_key}'",
            },
        )


def get_identity_types_for_connector(
    email_secrets: BaseEmailSchema,
) -> List[str]:
    """Return a list of identity types we need to email to the third party vendor."""
    advanced_settings: AdvancedSettings = email_secrets.advanced_settings
    identity_types: IdentityTypes = advanced_settings.identity_types
    flattened_list: List[str] = []

    if identity_types.email:
        flattened_list.append("email")
    if identity_types.phone_number:
        flattened_list.append("phone_number")

    return flattened_list


def filter_user_identities_for_connector(
    secrets: BaseEmailSchema, user_identities: Dict[str, Any]
) -> Dict[str, Any]:
    """Filter identities to just those specified for a given connector"""
    required_identities: List[str] = get_identity_types_for_connector(secrets)
    return {
        identity_type: user_identities.get(identity_type)
        for identity_type in required_identities
        if user_identities.get(identity_type)
    }


def send_single_erasure_email(
    db: Session,
    subject_email: str,
    subject_name: str,
    batch_identities: List[str],
    test_mode: bool = False,
) -> None:
    """Sends a single erasure email"""

    org_name = get_org_name(db)

    dispatch_message(
        db=db,
        action_type=MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT,
        to_identity=Identity(email=subject_email),
        service_type=get_email_messaging_config_service_type(db=db),
        message_body_params=ErasureRequestBodyParams(
            controller=org_name,
            third_party_vendor_name=subject_name,
            identities=batch_identities,
        ),
        subject_override=f"{'Test notification' if test_mode else 'Notification'} "
        f"of user erasure requests from {org_name}",
    )



"""privacy request approve

Revision ID: 906d7198df28
Revises: 5a966cd643d7
Create Date: 2022-03-11 19:49:26.450054

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "906d7198df28"
down_revision = "5a966cd643d7"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type privacyrequeststatus add value 'approved'")
    op.execute("alter type privacyrequeststatus add value 'denied'")

    op.add_column(
        "privacyrequest",
        sa.Column("reviewed_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column(
        "privacyrequest", sa.Column("reviewed_by", sa.String(), nullable=True)
    )
    op.create_foreign_key(
        "privacyrequest_reviewed_by_fkey",
        "privacyrequest",
        "fidesopsuser",
        ["reviewed_by"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.execute("delete from privacyrequest where status in ('approved', 'denied')")
    op.drop_constraint(
        "privacyrequest_reviewed_by_fkey", "privacyrequest", type_="foreignkey"
    )
    op.drop_column("privacyrequest", "reviewed_by")
    op.drop_column("privacyrequest", "reviewed_at")

    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('in_processing', 'complete', 'pending', 'error', 'paused')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")



"""remove saving preferences for system, vendor, and purpose

Remove these columns and any previous rows where consent was stored against these attributes.
You can no longer save preferences for systems, vendors, and purposes.

Revision ID: 81226042d7d4
Revises: 81886da90395
Create Date: 2023-10-02 00:34:46.828946

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text

revision = "81226042d7d4"
down_revision = "81886da90395"
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    # First let's clean up all rows in currentprivacypreference, lastservednotice, privacypreferencehistory, and servednoticehistory
    # where consent was saved or served against purposes, vendors, or systems
    bind.execute(
        text(
            "DELETE from currentprivacypreference where purpose is not null or vendor is not null or system is not null;"
        )
    )
    bind.execute(
        text(
            "DELETE from lastservednotice where purpose is not null or vendor is not null or system is not null;"
        )
    )
    bind.execute(
        text(
            "DELETE from privacypreferencehistory where purpose is not null or vendor is not null or system is not null;"
        )
    )
    bind.execute(
        text(
            "DELETE from servednoticehistory where purpose is not null or vendor is not null or system is not null;"
        )
    )

    # Drop unique constraints on identities x purpose/system/vendor fields for currentprivacypreference
    op.drop_constraint(
        "fides_user_device_identity_purpose", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_system", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_vendor", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint("identity_purpose", "currentprivacypreference", type_="unique")
    op.drop_constraint("identity_system", "currentprivacypreference", type_="unique")
    op.drop_constraint("identity_vendor", "currentprivacypreference", type_="unique")

    # Drop indices on purpose/system/vendor fields for currentprivacypreference

    op.drop_index(
        "ix_currentprivacypreference_purpose", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_system", table_name="currentprivacypreference"
    )
    op.drop_index(
        "ix_currentprivacypreference_vendor", table_name="currentprivacypreference"
    )
    # Now delete vendor/purpose/system columns on currentprivacypreference
    op.drop_column("currentprivacypreference", "vendor")
    op.drop_column("currentprivacypreference", "purpose")
    op.drop_column("currentprivacypreference", "system")

    # Remove index on purpose, system, and vendor in lastservednotice table
    op.drop_index("ix_lastservednotice_purpose", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_system", table_name="lastservednotice")
    op.drop_index("ix_lastservednotice_vendor", table_name="lastservednotice")

    # Drop unique constraints for identities x purpose/system/vendor on lastservednotice table
    op.drop_constraint(
        "last_served_fides_user_device_identity_purpose",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_system",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_vendor",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_identity_purpose", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_system", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_vendor", "lastservednotice", type_="unique"
    )
    # Now we can drop the columns for vendor/purpose/system on lastservednotice table
    op.drop_column("lastservednotice", "vendor")
    op.drop_column("lastservednotice", "purpose")
    op.drop_column("lastservednotice", "system")

    # Drop indices for purpose/system/vendor on privacypreferencehistory table
    op.drop_index(
        "ix_privacypreferencehistory_purpose", table_name="privacypreferencehistory"
    )
    op.drop_index(
        "ix_privacypreferencehistory_system", table_name="privacypreferencehistory"
    )
    op.drop_index(
        "ix_privacypreferencehistory_vendor", table_name="privacypreferencehistory"
    )
    # Now we can drop system/vendor/purpose columns on privacypreferencehistory
    op.drop_column("privacypreferencehistory", "system")
    op.drop_column("privacypreferencehistory", "purpose")
    op.drop_column("privacypreferencehistory", "vendor")

    # Drop indices for purpose/system/vendor on servednoticehistory table
    op.drop_index("ix_servednoticehistory_purpose", table_name="servednoticehistory")
    op.drop_index("ix_servednoticehistory_system", table_name="servednoticehistory")
    op.drop_index("ix_servednoticehistory_vendor", table_name="servednoticehistory")

    # Now we can drop system/purpose/vendor columns on servednoticehistory table
    op.drop_column("servednoticehistory", "system")
    op.drop_column("servednoticehistory", "purpose")
    op.drop_column("servednoticehistory", "vendor")


def downgrade():
    op.add_column(
        "servednoticehistory",
        sa.Column("vendor", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("system", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_index(
        "ix_servednoticehistory_vendor", "servednoticehistory", ["vendor"], unique=False
    )
    op.create_index(
        "ix_servednoticehistory_system", "servednoticehistory", ["system"], unique=False
    )
    op.create_index(
        "ix_servednoticehistory_purpose",
        "servednoticehistory",
        ["purpose"],
        unique=False,
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("vendor", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("system", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_index(
        "ix_privacypreferencehistory_vendor",
        "privacypreferencehistory",
        ["vendor"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_system",
        "privacypreferencehistory",
        ["system"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_purpose",
        "privacypreferencehistory",
        ["purpose"],
        unique=False,
    )
    op.add_column(
        "lastservednotice",
        sa.Column("system", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "lastservednotice",
        sa.Column("purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "lastservednotice",
        sa.Column("vendor", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_unique_constraint(
        "last_served_identity_vendor",
        "lastservednotice",
        ["provided_identity_id", "vendor"],
    )
    op.create_unique_constraint(
        "last_served_identity_system",
        "lastservednotice",
        ["provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "last_served_identity_purpose",
        "lastservednotice",
        ["provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_vendor",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "vendor"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_system",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_purpose",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "purpose"],
    )
    op.create_index(
        "ix_lastservednotice_vendor", "lastservednotice", ["vendor"], unique=False
    )
    op.create_index(
        "ix_lastservednotice_system", "lastservednotice", ["system"], unique=False
    )
    op.create_index(
        "ix_lastservednotice_purpose", "lastservednotice", ["purpose"], unique=False
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("system", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("vendor", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_index(
        "ix_currentprivacypreference_vendor",
        "currentprivacypreference",
        ["vendor"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_system",
        "currentprivacypreference",
        ["system"],
        unique=False,
    )
    op.create_index(
        "ix_currentprivacypreference_purpose",
        "currentprivacypreference",
        ["purpose"],
        unique=False,
    )
    op.create_unique_constraint(
        "identity_vendor",
        "currentprivacypreference",
        ["provided_identity_id", "vendor"],
    )
    op.create_unique_constraint(
        "identity_system",
        "currentprivacypreference",
        ["provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "identity_purpose",
        "currentprivacypreference",
        ["provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_vendor",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "vendor"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_system",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_purpose",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "purpose"],
    )



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import ConfigDict, Field, field_validator, model_validator

from fides.api.models.location_regulation_selections import PrivacyNoticeRegion
from fides.api.schemas.base_class import FidesSchema

RequiredType = Literal["optional", "required"]


class CustomIdentity(FidesSchema):
    label: str


class IdentityInputs(FidesSchema):
    name: Optional[RequiredType] = None
    email: Optional[RequiredType] = None
    phone: Optional[RequiredType] = None
    model_config = ConfigDict(extra="allow")

    def __init__(self, **data: Any):
        for field, value in data.items():
            if field not in self.model_fields:
                if isinstance(value, CustomIdentity):
                    data[field] = value
                elif isinstance(value, dict) and "label" in value:
                    data[field] = CustomIdentity(**value)
                else:
                    raise ValueError(
                        f'Custom identity "{field}" must be an instance of CustomIdentity '
                        '(e.g. {"label": "Field label"})'
                    )
        super().__init__(**data)


class CustomPrivacyRequestField(FidesSchema):
    label: str
    required: Optional[bool] = True
    default_value: Optional[str] = None
    hidden: Optional[bool] = False
    query_param_key: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def validate_default_value(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if (
            values.get("hidden")
            and values.get("default_value") is None
            and values.get("query_param_key") is None
        ):
            raise ValueError(
                "default_value or query_param_key are required when hidden is True"
            )
        return values


class PrivacyRequestOption(FidesSchema):
    locations: Optional[Union[List[PrivacyNoticeRegion], Literal["fallback"]]] = None
    policy_key: Optional[str] = None
    icon_path: str
    title: str
    description: str
    description_subtext: Optional[List[str]] = None
    confirm_button_text: Optional[str] = Field(alias="confirmButtonText", default=None)
    cancel_button_text: Optional[str] = Field(alias="cancelButtonText", default=None)
    identity_inputs: Optional[IdentityInputs] = None
    custom_privacy_request_fields: Optional[Dict[str, CustomPrivacyRequestField]] = None


class ConsentConfigButton(FidesSchema):
    description: str
    description_subtext: Optional[List[str]] = None
    confirm_button_text: Optional[str] = Field(alias="confirmButtonText", default=None)
    cancel_button_text: Optional[str] = Field(alias="cancelButtonText", default=None)
    icon_path: str
    identity_inputs: IdentityInputs
    custom_privacy_request_fields: Optional[Dict[str, CustomPrivacyRequestField]] = None
    title: str
    modal_title: Optional[str] = Field(alias="modalTitle", default=None)


class ConditionalValue(FidesSchema):
    value: bool
    global_privacy_control: bool = Field(alias="globalPrivacyControl")


class ConfigConsentOption(FidesSchema):
    cookie_keys: List[str] = Field([], alias="cookieKeys")
    default: Optional[Union[bool, ConditionalValue]] = None
    description: str
    fides_data_use_key: str = Field(alias="fidesDataUseKey")
    highlight: Optional[bool] = None
    name: str
    url: str
    executable: Optional[bool] = None


class ConsentConfigPage(FidesSchema):
    consent_options: List[ConfigConsentOption] = Field([], alias="consentOptions")
    description: str
    description_subtext: Optional[List[str]] = None
    policy_key: Optional[str] = None
    title: str

    @field_validator("consent_options")
    @classmethod
    def validate_consent_options(
        cls, consent_options: List[ConfigConsentOption]
    ) -> List[ConfigConsentOption]:
        executable_count = sum(option.executable is True for option in consent_options)
        if executable_count > 1:
            raise ValueError("Cannot have more than one consent option be executable.")
        return consent_options


class ConsentConfig(FidesSchema):
    button: ConsentConfigButton
    page: ConsentConfigPage


class PolicyUnavailableMessages(FidesSchema):
    """
    Used to capture the information to present to a user if a policy is unavailable.
    """

    title: str
    description: str
    close_button_text: str
    action_button_text: str
    action_link: str


class PrivacyCenterConfig(FidesSchema):
    """
    NOTE: Add to this schema with care. Any fields added to
    this response schema will be exposed in public-facing
    (i.e. unauthenticated) API responses. If a field has
    sensitive information, it should NOT be added to this schema!
    """

    title: str
    description: str
    description_subtext: Optional[List[str]] = None
    addendum: Optional[List[str]] = None
    server_url_development: Optional[str] = None
    server_url_production: Optional[str] = None
    logo_path: Optional[str] = None
    logo_url: Optional[str] = None
    favicon_path: Optional[str] = None
    actions: List[PrivacyRequestOption]
    include_consent: Optional[bool] = Field(alias="includeConsent", default=None)
    consent: ConsentConfig
    privacy_policy_url: Optional[str] = None
    privacy_policy_url_text: Optional[str] = None
    policy_unavailable_messages: Optional[PolicyUnavailableMessages] = None


class PartialPrivacyRequestOption(FidesSchema):
    policy_key: str
    title: str
    identity_inputs: Optional[IdentityInputs] = None
    custom_privacy_request_fields: Optional[Dict[str, CustomPrivacyRequestField]] = None


class PartialPrivacyCenterConfig(FidesSchema):
    """Partial schema for the Admin UI privacy request submission."""

    actions: List[PartialPrivacyRequestOption]



"""add system vendor_deleted_date field

Revision ID: d873e3e430b0
Revises: 23e01f8f4652
Create Date: 2024-04-22 14:25:13.232408

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d873e3e430b0"
down_revision = "23e01f8f4652"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems",
        sa.Column("vendor_deleted_date", sa.DateTime(timezone=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ctl_systems", "vendor_deleted_date")
    # ### end Alembic commands ###



"""add external id to privacy preference

Revision ID: 5fe01e730171
Revises: 5f96f13be8ec
Create Date: 2024-05-31 17:11:50.566937

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "5fe01e730171"
down_revision = "5f96f13be8ec"
branch_labels = None
depends_on = None

"""
WARNING - Conditional migration based on table size

This migration script checks the size of specific tables and performs the following actions:

1. If the tables have less than 1 million rows:
   - Creates the necessary indices
   - Adds the required constraints

2. If the tables are too large (1 million rows or more):
   - Logs non-blocking SQL commands to create the indices and constraints
   - The non-blocking commands utilize:
     - `CREATE INDEX CONCURRENTLY` for creating indices
     - `ADD CONSTRAINT USING INDEX` for adding constraints
   - These commands add the indices and constraints without blocking any reads or writes on the tables
"""


def upgrade():
    op.add_column(
        "consentrequest",
        sa.Column("property_id", sa.String(), nullable=True),
    )
    op.add_column(
        "currentprivacypreferencev2",
        sa.Column(
            "external_id",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "currentprivacypreferencev2",
        sa.Column("hashed_external_id", sa.String(), nullable=True),
    )
    op.add_column(
        "currentprivacypreferencev2",
        sa.Column("property_id", sa.String(), nullable=True),
    )
    op.drop_constraint(
        "last_saved_for_email", "currentprivacypreferencev2", type_="unique"
    )
    op.drop_constraint(
        "last_saved_for_fides_user_device", "currentprivacypreferencev2", type_="unique"
    )
    op.drop_constraint(
        "last_saved_for_phone_number", "currentprivacypreferencev2", type_="unique"
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "external_id",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("hashed_external_id", sa.String(), nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "external_id",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("hashed_external_id", sa.String(), nullable=True),
    )

    connection = op.get_bind()

    currentprivacypreferencev2_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM currentprivacypreferencev2")
    ).scalar()
    if currentprivacypreferencev2_count < 1000000:
        op.create_index(
            op.f("ix_currentprivacypreferencev2_hashed_external_id"),
            "currentprivacypreferencev2",
            ["hashed_external_id"],
            unique=False,
        )
        op.create_index(
            op.f("ix_currentprivacypreferencev2_email_property_id"),
            "currentprivacypreferencev2",
            ["email", "property_id"],
            unique=True,
        )
        op.create_index(
            op.f("ix_currentprivacypreferencev2_external_id_property_id"),
            "currentprivacypreferencev2",
            ["external_id", "property_id"],
            unique=True,
        )
        op.create_index(
            op.f("ix_currentprivacypreferencev2_fides_user_device_property_id"),
            "currentprivacypreferencev2",
            ["fides_user_device", "property_id"],
            unique=True,
        )
        op.create_index(
            op.f("ix_currentprivacypreferencev2_phone_number_property_id"),
            "currentprivacypreferencev2",
            ["phone_number", "property_id"],
            unique=True,
        )
        op.execute(
            """
            ALTER TABLE currentprivacypreferencev2
            ADD CONSTRAINT last_saved_for_email_per_property_id
            UNIQUE USING INDEX ix_currentprivacypreferencev2_email_property_id;
            """
        )
        op.execute(
            """
            ALTER TABLE currentprivacypreferencev2
            ADD CONSTRAINT last_saved_for_external_id_per_property_id
            UNIQUE USING INDEX ix_currentprivacypreferencev2_external_id_property_id;
            """
        )
        op.execute(
            """
            ALTER TABLE currentprivacypreferencev2
            ADD CONSTRAINT last_saved_for_fides_user_device_per_property_id
            UNIQUE USING INDEX ix_currentprivacypreferencev2_fides_user_device_property_id;
            """
        )
        op.execute(
            """
            ALTER TABLE currentprivacypreferencev2
            ADD CONSTRAINT last_saved_for_phone_number_per_property_id
            UNIQUE USING INDEX ix_currentprivacypreferencev2_phone_number_property_id;
            """
        )
    else:
        logger.warning(
            "The currentprivacypreferencev2 table has more than 1 million rows, "
            "skipping index and constraint creation. Be sure to manually run the following commands:\n"
            "- 'CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_email_property_id "
            "ON currentprivacypreferencev2 (email, property_id)'\n"
            "- 'CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_external_id_property_id "
            "ON currentprivacypreferencev2 (external_id, property_id)'\n"
            "- 'CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_fides_user_device_property_id "
            "ON currentprivacypreferencev2 (fides_user_device, property_id)'\n"
            "- 'CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_phone_number_property_id "
            "ON currentprivacypreferencev2 (phone_number, property_id)'\n"
            "- 'CREATE INDEX CONCURRENTLY ix_currentprivacypreferencev2_hashed_external_id "
            "ON currentprivacypreferencev2 (hashed_external_id)'\n"
            "- 'ALTER TABLE currentprivacypreferencev2 "
            "ADD CONSTRAINT last_saved_for_email_per_property_id "
            "UNIQUE USING INDEX ix_currentprivacypreferencev2_email_property_id'\n"
            "- 'ALTER TABLE currentprivacypreferencev2 "
            "ADD CONSTRAINT last_saved_for_external_id_per_property_id "
            "UNIQUE USING INDEX ix_currentprivacypreferencev2_external_id_property_id'\n"
            "- 'ALTER TABLE currentprivacypreferencev2 "
            "ADD CONSTRAINT last_saved_for_fides_user_device_per_property_id "
            "UNIQUE USING INDEX ix_currentprivacypreferencev2_fides_user_device_property_id'\n"
            "- 'ALTER TABLE currentprivacypreferencev2 "
            "ADD CONSTRAINT last_saved_for_phone_number_per_property_id "
            "UNIQUE USING INDEX ix_currentprivacypreferencev2_phone_number_property_id'"
        )

    privacypreferencehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM privacypreferencehistory")
    ).scalar()

    if privacypreferencehistory_count < 1000000:
        op.create_index(
            op.f("ix_privacypreferencehistory_hashed_external_id"),
            "privacypreferencehistory",
            ["hashed_external_id"],
            unique=False,
        )
    else:
        logger.warning(
            "The privacypreferencehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY ix_privacypreferencehistory_hashed_external_id "
            "ON privacypreferencehistory (hashed_external_id)'"
        )

    servednoticehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM servednoticehistory")
    ).scalar()
    if servednoticehistory_count < 1000000:
        op.create_index(
            op.f("ix_servednoticehistory_hashed_external_id"),
            "servednoticehistory",
            ["hashed_external_id"],
            unique=False,
        )
    else:
        logger.warning(
            "The servednoticehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY ix_servednoticehistory_hashed_external_id "
            "ON servednoticehistory (hashed_external_id)'"
        )


def downgrade():
    op.drop_index(
        op.f("ix_servednoticehistory_hashed_external_id"),
        table_name="servednoticehistory",
    )
    op.drop_column("servednoticehistory", "hashed_external_id")
    op.drop_column("servednoticehistory", "external_id")
    op.drop_index(
        op.f("ix_privacypreferencehistory_hashed_external_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "hashed_external_id")
    op.drop_column("privacypreferencehistory", "external_id")
    op.drop_constraint(
        "last_saved_for_phone_number_per_property_id",
        "currentprivacypreferencev2",
        type_="unique",
    )
    op.drop_constraint(
        "last_saved_for_fides_user_device_per_property_id",
        "currentprivacypreferencev2",
        type_="unique",
    )
    op.drop_constraint(
        "last_saved_for_external_id_per_property_id",
        "currentprivacypreferencev2",
        type_="unique",
    )
    op.drop_constraint(
        "last_saved_for_email_per_property_id",
        "currentprivacypreferencev2",
        type_="unique",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_property_id"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_hashed_external_id"),
        table_name="currentprivacypreferencev2",
    )
    op.create_unique_constraint(
        "last_saved_for_phone_number", "currentprivacypreferencev2", ["phone_number"]
    )
    op.create_unique_constraint(
        "last_saved_for_fides_user_device",
        "currentprivacypreferencev2",
        ["fides_user_device"],
    )
    op.create_unique_constraint(
        "last_saved_for_email", "currentprivacypreferencev2", ["email"]
    )
    op.drop_column("currentprivacypreferencev2", "property_id")
    op.drop_column("currentprivacypreferencev2", "hashed_external_id")
    op.drop_column("currentprivacypreferencev2", "external_id")
    op.drop_column("consentrequest", "property_id")



import importlib
import os
import os.path

# path to the current directory
directory = os.path.dirname(__file__)

# loop through the files in the override_implementations directory
for filename in os.listdir(directory):
    # ignore non-Python files and the __init__.py file
    if filename.endswith(".py") and filename != "__init__.py":
        # import the module
        module = importlib.import_module(f"{__name__}.{filename[:-3]}")



import uuid
from datetime import datetime
from typing import Optional, Tuple

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.api.v1.endpoints.messaging_endpoints import user_email_invite_status
from fides.api.common_exceptions import AuthorizationError
from fides.api.models.client import ClientDetail
from fides.api.models.fides_user import FidesUser
from fides.api.models.fides_user_invite import FidesUserInvite
from fides.api.schemas.messaging.messaging import (
    MessagingActionType,
    UserInviteBodyParams,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.messaging.message_dispatch_service import dispatch_message
from fides.config import FidesConfig
from fides.config.config_proxy import ConfigProxy


def invite_user(db: Session, config_proxy: ConfigProxy, user: FidesUser) -> None:
    """
    Generates a user invite and sends the invite code to the user via email.

    This is a no-op if email messaging isn't configured.
    """

    # invite user via email if email messaging is enabled and the Admin UI URL is defined
    if user_email_invite_status(db=db, config_proxy=config_proxy).enabled:
        invite_code = str(uuid.uuid4())
        FidesUserInvite.create(
            db=db, data={"username": user.username, "invite_code": invite_code}
        )
        user.update(db, data={"disabled": True})
        dispatch_message(
            db,
            action_type=MessagingActionType.USER_INVITE,
            to_identity=Identity(email=user.email_address),
            service_type=config_proxy.notifications.notification_service_type,
            message_body_params=UserInviteBodyParams(
                username=user.username, invite_code=invite_code
            ),
        )


def accept_invite(
    db: Session, config: FidesConfig, user: FidesUser, new_password: str
) -> Tuple[FidesUser, str]:
    """
    Updates the user password and enables the user. Also removes the user invite from the database.
    Returns a tuple of the updated user and their access code.
    """

    # update password and enable
    user.update_password(db=db, new_password=new_password)
    user.update(
        db,
        data={"disabled": False, "disabled_reason": None},
    )
    db.refresh(user)

    # delete invite
    if user.username:
        invite = FidesUserInvite.get_by(db=db, field="username", value=user.username)
        if invite:
            invite.delete(db)
    else:
        logger.warning("Username is missing, skipping invite deletion.")

    client = perform_login(
        db,
        config.security.oauth_client_id_length_bytes,
        config.security.oauth_client_secret_length_bytes,
        user,
    )

    logger.info("Creating login access token")
    access_code = client.create_access_code_jwe(config.security.app_encryption_key)

    return user, access_code


def perform_login(
    db: Session,
    client_id_byte_length: int,
    client_secret_byte_length: int,
    user: FidesUser,
    skip_save: Optional[bool] = False,
) -> ClientDetail:
    """Performs a login by updating the FidesUser instance and creating and returning
    an associated ClientDetail.

    If the username or password was bad, skip_save should be True. We still run through
    parallel operations to keep the timing of operations similar, but should skip
    saving to the database.
    """

    client = user.client
    if not client:
        logger.info("Creating client for login")
        client, _ = ClientDetail.create_client_and_secret(
            db,
            client_id_byte_length,
            client_secret_byte_length,
            scopes=[],  # type: ignore
            roles=user.permissions.roles,  # type: ignore
            systems=user.system_ids,  # type: ignore
            user_id=user.id,
            in_memory=skip_save,  # If login flow has already errored, don't persist this to the database
        )
    else:
        # Refresh the client just in case - for example, scopes and roles were added via the db directly.
        client.roles = user.permissions.roles  # type: ignore
        client.systems = user.system_ids  # type: ignore
        if not skip_save:
            client.save(db)

    if user.permissions and (not user.permissions.roles and not user.systems):  # type: ignore
        logger.warning("User {} needs roles or systems to login.", user.id)
        raise AuthorizationError(detail="Not Authorized for this action")

    if not skip_save:
        user.last_login_at = datetime.utcnow()
        user.save(db)

    return client



"""This module defines the settings for everything related to the CLI."""

from typing import Optional
from urllib.parse import urlparse

from fideslang.validation import AnyHttpUrlString
from fideslog.sdk.python.utils import FIDESCTL_CLI, generate_client_id
from pydantic import Field, SerializeAsAny, ValidationInfo, field_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings, port_integer_converter

# pylint: disable=C0115,C0116, E0213

ENV_PREFIX = "FIDES__CLI__"


class CLISettings(FidesSettings):
    """Configuration settings for the command-line application."""

    analytics_id: str = Field(
        default=generate_client_id(FIDESCTL_CLI),
        description="A fully anonymized unique identifier that is automatically generated by the application. Used for anonymous analytics when opted-in.",
        validate_default=True,
    )
    local_mode: bool = Field(
        default=False,
        description="When set to True, disables functionality that requires making calls to a Fides webserver.",
    )
    server_protocol: str = Field(
        default="http", description="The protocol used by the Fides webserver."
    )
    server_host: str = Field(
        default="localhost", description="The hostname of the Fides webserver."
    )
    server_port: str = Field(
        default="8080", description="The port of the Fides webserver"
    )
    server_path: str = Field(default="/", description="The path of the Fides webserver")
    server_url: SerializeAsAny[Optional[AnyHttpUrlString]] = Field(
        default=None,
        description="The full server url generated from the other server configuration values.",
        exclude=True,
        validate_default=True,
    )

    @field_validator("server_url")
    @classmethod
    def get_server_url(cls, value: str, info: ValidationInfo) -> str:
        """Create the server_url from the server configuration values. Strips path if present."""
        host = info.data.get("server_host")
        protocol = info.data.get("server_protocol")
        port: int = port_integer_converter(info, "server_port")
        port_str: str = f":{port}" if port else ""
        server_url = f"{protocol}://{host}{port_str}"
        # check if the host var had a path included
        parsed = urlparse(server_url)
        if parsed.path:
            print(
                f"Warning: The server_host value '{host}' includes a path. This will be stripped. Please update the server_host value to '{parsed.netloc}'."
            )
            server_url = f"{parsed.scheme}://{parsed.netloc}{port_str}"
            print(f"The server_url: `{server_url}` will be used.")

        return server_url

    @field_validator("analytics_id")
    def ensure_not_empty(cls, value: str) -> str:
        """
        Validate that the `analytics_id` is not `""`.
        """
        return value if value != "" else generate_client_id(FIDESCTL_CLI)

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX, coerce_numbers_to_str=True)



import datetime
from typing import Annotated, Dict, List, Optional, Union

from fastapi import Depends, HTTPException, Query, Response, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.async_sqlalchemy import paginate as async_paginate
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.models import System as SystemSchema
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from sqlalchemy import or_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import Session
from starlette import status
from starlette.status import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND

from fides.api.api import deps
from fides.api.api.v1.endpoints.saas_config_endpoints import instantiate_connection
from fides.api.db.crud import (
    get_resource,
    get_resource_with_custom_fields,
    list_resource,
)
from fides.api.db.ctl_session import get_async_db
from fides.api.db.system import (
    create_system,
    get_system,
    update_system,
    upsert_system,
    validate_privacy_declarations,
)
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.models.fides_user import FidesUser
from fides.api.models.sql_models import (  # type:ignore[attr-defined]
    PrivacyDeclaration,
    System,
)
from fides.api.oauth.system_manager_oauth_util import (
    verify_oauth_client_for_system_from_fides_key,
    verify_oauth_client_for_system_from_request_body_cli,
)
from fides.api.oauth.utils import get_current_user, verify_oauth_client_prod
from fides.api.schemas.connection_configuration import connection_secrets_schemas
from fides.api.schemas.connection_configuration.connection_config import (
    BulkPutConnectionConfiguration,
    ConnectionConfigurationResponse,
    CreateConnectionConfigurationWithSecrets,
    SaasConnectionTemplateResponse,
)
from fides.api.schemas.connection_configuration.connection_secrets import (
    TestStatusMessage,
)
from fides.api.schemas.connection_configuration.saas_config_template_values import (
    SaasConnectionTemplateValues,
)
from fides.api.schemas.filter_params import FilterParams
from fides.api.schemas.system import BasicSystemResponse, SystemResponse
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_util import (
    connection_status,
    delete_connection_config,
    get_connection_config_or_error,
    patch_connection_configs,
    validate_secrets,
)
from fides.api.util.filter_utils import apply_filters_to_query
from fides.common.api.scope_registry import (
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_DELETE,
    CONNECTION_READ,
    SAAS_CONNECTION_INSTANTIATE,
    SYSTEM_CREATE,
    SYSTEM_DELETE,
    SYSTEM_READ,
    SYSTEM_UPDATE,
)
from fides.common.api.v1.urn_registry import (
    INSTANTIATE_SYSTEM_CONNECTION,
    SYSTEM_CONNECTIONS,
    V1_URL_PREFIX,
)

SYSTEM_ROUTER = APIRouter(tags=["System"], prefix=f"{V1_URL_PREFIX}/system")
SYSTEM_CONNECTIONS_ROUTER = APIRouter(
    tags=["System"], prefix=f"{V1_URL_PREFIX}{SYSTEM_CONNECTIONS}"
)
SYSTEM_CONNECTION_INSTANTIATE_ROUTER = APIRouter(
    tags=["System"],
    prefix=f"{V1_URL_PREFIX}{INSTANTIATE_SYSTEM_CONNECTION}",
)


@SYSTEM_CONNECTIONS_ROUTER.get(
    "",
    dependencies=[
        Security(
            verify_oauth_client_for_system_from_fides_key, scopes=[CONNECTION_READ]
        )
    ],
    status_code=HTTP_200_OK,
    response_model=Page[ConnectionConfigurationResponse],
)
def get_system_connections(
    fides_key: str, params: Params = Depends(), db: Session = Depends(deps.get_db)
) -> AbstractPage[ConnectionConfigurationResponse]:
    """
    Return all the connection configs related to a system.
    """
    system = get_system(db, fides_key)
    query = ConnectionConfig.query(db)
    query = query.filter(ConnectionConfig.system_id == system.id)
    return paginate(query.order_by(ConnectionConfig.name.asc()), params=params)


@SYSTEM_CONNECTIONS_ROUTER.patch(
    "",
    dependencies=[
        Security(
            verify_oauth_client_for_system_from_fides_key,
            scopes=[CONNECTION_CREATE_OR_UPDATE],
        )
    ],
    status_code=HTTP_200_OK,
    response_model=BulkPutConnectionConfiguration,
)
def patch_connections(
    fides_key: str,
    configs: Annotated[List[CreateConnectionConfigurationWithSecrets], Field(max_length=50)],  # type: ignore
    db: Session = Depends(deps.get_db),
) -> BulkPutConnectionConfiguration:
    """
    Given a valid System fides key, a list of connection config data elements, optionally
    containing the secrets, create or update corresponding ConnectionConfig objects or report
    failure.

    If the key in the payload exists, it will be used to update an existing ConnectionConfiguration.
    Otherwise, a new ConnectionConfiguration will be created for you.
    """
    system = get_system(db, fides_key)
    return patch_connection_configs(db, configs, system)


@SYSTEM_CONNECTIONS_ROUTER.patch(
    "/secrets",
    dependencies=[
        Security(
            verify_oauth_client_for_system_from_fides_key,
            scopes=[CONNECTION_CREATE_OR_UPDATE],
        )
    ],
    status_code=HTTP_200_OK,
    response_model=TestStatusMessage,
)
def patch_connection_secrets(
    fides_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_secrets: connection_secrets_schemas,
    verify: Optional[bool] = True,
) -> TestStatusMessage:
    """
    Patch secrets that will be used to connect to a specified connection_type.

    The specific secrets will be connection-dependent. For example, the components needed to connect to a Postgres DB
    will differ from Dynamo DB.
    """

    system = get_system(db, fides_key)
    connection_config: ConnectionConfig = get_connection_config_or_error(
        db, system.connection_configs.key
    )
    # Inserts unchanged sensitive values. The FE does not send masked values sensitive secrets.
    if connection_config.secrets is not None:
        for key, value in connection_config.secrets.items():
            if key not in unvalidated_secrets:
                # unvalidated_secrets is actually a dictionary here.  connection_secrets_schemas
                # are just provided for documentation but the data was not parsed up front.
                # That happens below in validate_secrets.
                unvalidated_secrets[key] = value  # type: ignore
    else:
        connection_config.secrets = {}

    validated_secrets = validate_secrets(
        db, unvalidated_secrets, connection_config
    ).model_dump(mode="json")

    for key, value in validated_secrets.items():
        connection_config.secrets[key] = value  # type: ignore

    # Deauthorize an OAuth connection when the secrets are updated. This is necessary because
    # the existing access tokens may not be valid anymore. This only applies to SaaS connection
    # configurations that use the "oauth2_authorization_code" authentication strategy.
    if (
        connection_config.authorized
        and connection_config.connection_type == ConnectionType.saas
    ):
        del connection_config.secrets["access_token"]

    # Save validated secrets, regardless of whether they've been verified.
    logger.info("Updating connection config secrets for '{}'", connection_config.key)
    connection_config.save(db=db)

    msg = f"Secrets updated for ConnectionConfig with key: {connection_config.key}."
    if verify:
        return connection_status(connection_config, msg, db)

    return TestStatusMessage(msg=msg, test_status=None)


@SYSTEM_CONNECTIONS_ROUTER.delete(
    "",
    dependencies=[
        Security(
            verify_oauth_client_for_system_from_fides_key, scopes=[CONNECTION_DELETE]
        )
    ],
    status_code=HTTP_204_NO_CONTENT,
    response_model=None,
)
def delete_connection(fides_key: str, *, db: Session = Depends(deps.get_db)) -> None:
    system = get_system(db, fides_key)
    if system.connection_configs is None:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No integration found linked to this system",
        )

    # system.connection_configs will temporarily only have one config
    # it will be updated to have multiple configs in the future
    delete_connection_config(db, system.connection_configs.key)


@SYSTEM_ROUTER.put(
    "/",
    response_model=SystemResponse,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "error": "user does not have permission to modify this resource",
                            "resource_type": "system",
                            "fides_key": "example.key",
                        }
                    }
                }
            }
        },
    },
)
async def update(
    resource: SystemSchema = Security(
        verify_oauth_client_for_system_from_request_body_cli,
        scopes=[SYSTEM_UPDATE],
    ),  # Security dependency defined here instead of the path operation decorator so we have access to the request body
    # to be able to look up the system as well as return a value
    db: AsyncSession = Depends(get_async_db),
    current_user: FidesUser = Depends(get_current_user),
) -> Dict:
    """
    Update a System by the fides_key extracted from the request body.  Defined outside of the crud routes
    to add additional "system manager" permission checks.
    """
    await validate_privacy_declarations(db, resource)
    updated_system, _ = await update_system(
        resource, db, current_user.id if current_user else None
    )
    return updated_system


@SYSTEM_ROUTER.post(
    "/upsert",
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[
                SYSTEM_CREATE,
                SYSTEM_UPDATE,
            ],
        )
    ],
)
async def upsert(
    resources: List[SystemSchema],
    response: Response,
    db: AsyncSession = Depends(get_async_db),
    current_user: FidesUser = Depends(get_current_user),
) -> Dict:
    inserted, updated = await upsert_system(
        resources, db, current_user.id if current_user else None
    )
    response.status_code = (
        status.HTTP_201_CREATED if inserted > 0 else response.status_code
    )
    return {
        "message": f"Upserted {len(resources)} System(s)",
        "inserted": inserted,
        "updated": updated,
    }


@SYSTEM_ROUTER.delete(
    "/{fides_key}",
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[SYSTEM_DELETE],
        )
    ],
    responses={
        status.HTTP_403_FORBIDDEN: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "error": "user does not have permission to modify this resource",
                            "resource_type": "system",
                            "fides_key": "example.key",
                        }
                    }
                }
            }
        },
    },
)
async def delete(
    fides_key: str,
    # to retrieve the System and also return a value
    db: AsyncSession = Depends(get_async_db),
) -> Dict:
    """
    Delete a System by its fides_key. Defined outside of the crud routes
    to add additional "system manager" permission checks.
    """
    system_to_delete = await get_resource(System, fides_key, db)
    async with db.begin():
        await db.delete(system_to_delete)
    # Convert the resource to a dict explicitly for the response
    deleted_resource_dict = SystemSchema.model_validate(system_to_delete).model_dump(
        mode="json"
    )
    return {
        "message": "resource deleted",
        "resource": deleted_resource_dict,
    }


@SYSTEM_ROUTER.post(
    "/",
    response_model=SystemResponse,
    status_code=status.HTTP_201_CREATED,
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[SYSTEM_CREATE],
        )
    ],
)
async def create(
    resource: SystemSchema,
    db: AsyncSession = Depends(get_async_db),
    current_user: FidesUser = Depends(get_current_user),
) -> Dict:
    """
    Override `System` create/POST to handle `.privacy_declarations` defined inline,
    for backward compatibility and ease of use for API users.
    """
    return await create_system(resource, db, current_user.id if current_user else None)


@SYSTEM_ROUTER.get(
    "/",
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[SYSTEM_READ],
        )
    ],
    response_model=Union[List[BasicSystemResponse], Page[BasicSystemResponse]],
    name="List systems (optionally paginated)",
)
async def ls(  # pylint: disable=invalid-name
    db: AsyncSession = Depends(get_async_db),
    size: Optional[int] = Query(None, ge=1, le=100),
    page: Optional[int] = Query(None, ge=1),
    search: Optional[str] = None,
    data_uses: Optional[List[FidesKey]] = Query(None),
    data_categories: Optional[List[FidesKey]] = Query(None),
    data_subjects: Optional[List[FidesKey]] = Query(None),
    show_deleted: Optional[bool] = Query(False),
) -> List:
    """Get a list of all of the Systems.
    If any parameters or filters are provided the response will be paginated and/or filtered.
    Otherwise all Systems will be returned (this may be a slow operation if there are many systems,
    so using the pagination parameters is recommended).
    """
    if not (size or page or search or data_uses or data_categories or data_subjects):
        # if no advanced parameters are passed, we return a very basic list of all System resources
        # to maintain backward compatibility of the original API, which backs some important client usages, e.g. the fides CLI

        return await list_resource(System, db)

    query = select(System)

    pagination_params = Params(page=page or 1, size=size or 50)
    # Need to join with PrivacyDeclaration in order to be able to filter
    # by data use, data category, and data subject
    if any([data_uses, data_categories, data_subjects]):
        query = query.outerjoin(
            PrivacyDeclaration, System.id == PrivacyDeclaration.system_id
        )

    # Filter out any vendor deleted systems, unless explicitly asked for
    if not show_deleted:
        query = query.filter(
            or_(
                System.vendor_deleted_date.is_(None),
                System.vendor_deleted_date >= datetime.datetime.now(),
            )
        )

    filter_params = FilterParams(
        search=search,
        data_uses=data_uses,
        data_categories=data_categories,
        data_subjects=data_subjects,
    )
    filtered_query = apply_filters_to_query(
        query=query,
        filter_params=filter_params,
        search_model=System,
        taxonomy_model=PrivacyDeclaration,
    )

    # Add a distinct so we only get one row per system
    duplicates_removed = filtered_query.distinct(System.id)

    return await async_paginate(db, duplicates_removed, pagination_params)


@SYSTEM_ROUTER.patch(
    "/hidden",
    response_model=Dict,
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[SYSTEM_UPDATE],
        )
    ],
)
def patch_hidden(
    fides_keys: List[str],
    hidden: bool,
    db: Session = Depends(deps.get_db),
) -> Dict:
    """
    Patch the hidden status of a list of systems. Request body must be a list of system Fides keys.
    """
    systems = db.execute(select(System).filter(System.fides_key.in_(fides_keys)))
    systems = systems.scalars().all()

    for system in systems:
        system.hidden = hidden
    db.commit()

    return {
        "message": "Updated hidden status for systems",
        "updated": len(systems),
    }


@SYSTEM_ROUTER.get(
    "/{fides_key}",
    dependencies=[
        Security(
            verify_oauth_client_prod,
            scopes=[SYSTEM_READ],
        )
    ],
    response_model=SystemResponse,
)
async def get(
    fides_key: str,
    db: AsyncSession = Depends(get_async_db),
) -> Dict:
    """Get a resource by its fides_key."""
    return await get_resource_with_custom_fields(System, fides_key, db)


@SYSTEM_CONNECTION_INSTANTIATE_ROUTER.post(
    "/",
    dependencies=[
        Security(verify_oauth_client_prod, scopes=[SAAS_CONNECTION_INSTANTIATE])
    ],
    response_model=SaasConnectionTemplateResponse,
)
def instantiate_connection_from_template(
    fides_key: str,
    saas_connector_type: str,
    template_values: SaasConnectionTemplateValues,
    db: Session = Depends(deps.get_db),
) -> SaasConnectionTemplateResponse:
    """
    Instantiates a SaaS connection from the available template and the template values.
    Associates the newly instantiated connection with the provided system.
    """

    system = get_system(db, fides_key)
    return instantiate_connection(db, saas_connector_type, template_values, system)



"""Add created_at and updated_at

Revision ID: 7c851d8a102a
Revises: 45c7a349db68
Create Date: 2021-12-16 12:20:03.801089

"""

from alembic import op
from sqlalchemy import Column, DateTime, text

from fides.api.models.sql_models import sql_model_map

# revision identifiers, used by Alembic.
revision = "7c851d8a102a"
down_revision = "312aff72b275"
branch_labels = None
depends_on = None

SQL_MODEL_LIST = [
    "data_categories",
    "data_qualifiers",
    "data_uses",
    "data_subjects",
    "datasets",
    "evaluations",
    "policies",
    "registries",
    "systems",
    "organizations",
]


def upgrade():
    for table_name in SQL_MODEL_LIST:
        op.add_column(
            table_name,
            Column(
                "created_at",
                DateTime(timezone=True),
                server_default=text("now()"),
                nullable=True,
            ),
        )
        op.add_column(
            table_name,
            Column(
                "updated_at",
                DateTime(timezone=True),
                server_default=text("now()"),
                nullable=True,
            ),
        )


def downgrade():
    for table_name in SQL_MODEL_LIST:
        op.drop_column(table_name, "created_at")
        op.drop_column(table_name, "updated_at")
